#!/usr/bin/python2.7
#-*- coding: utf-8 -*-
from __future__ import print_function


import sys
sys.path.insert(0, 'lib')
from InstLC3 import *
from LibGraphes import *


import networkx as nx
import graphviz as gz
import copy
import itertools


debug=True

def RepresentsImm(s):
	try: 
		int(s)
		return True
	except ValueError:
		try:
			int(s[1:]) #for sharp+num
			return True
		except ValueError:
			return False

#t1 and t2 interfere iff
def interfere(t1,t2,mapout,defined):
	# TODO
	raise Exception("Not Yet Implemented")
				
#CLASS FOR A BLOCK (1 statement per block)
				
class Block:
	count = 0
	def __init__(self):
		self._i = Block.count
		self._in = []
		self._out = []
		self._ins = None
		self._left = None
		self._right = []
		Block.count += 1
		#for liveness dataflow analysis
		self._gen = set()
		self._kill = set()

				
	def append(self, child):
		self._out.append(child)
		child._in.append(self)
		return child

	def checkDep(self, visited):
		if self in visited:
			return False
		if self._in:
			for b in self._in:
				if b not in visited and b._i < self._i:
					return False
			return True
		else:
			return True

	def printIns(self, stream, visited,alloc):
		if self.checkDep(visited):
			visited.add(self)
			if self._ins:
				self._ins.printIns(stream,alloc)
			if self._out:
				for child in self._out:
					child.printIns(stream, visited,alloc)
		else:
			pass

		#Prints the current block and its sons.
	def printDot(self, graph, visited):
		if self.checkDep(visited):
			visited.add(self)
			if self._ins:
				if self._left:
					s1 = str(self._i)+'_'+self._ins.ins+' ('+self._left+' <- '+str(self._right)+')'
					graph.add_node(s1)
				else:
					s1 = str(self._i)+'_'+self._ins.ins+' (_ <- '+str(self._right)+')'
					graph.add_node(s1)
			else:
				s1 = "START"
				graph.add_node(s1)
			
			if self._out:
				for child in self._out:
					if child._left:
						s2 = str(child._i)+'_'+child._ins.ins+' ('+child._left+' <- '+str(child._right)+')'
					else:
						s2 = str(child._i)+'_'+child._ins.ins+' (_ <- '+str(child._right)+')'
					graph.add_edge(s1,s2)
					child.printDot(graph, visited)
		else:
			pass

	#propagate dataflow information and update LV_in and LV_out sets (mapin,mapout)
	#and make a recursive call to my sons.
	def do_dataflow_onestep(self, mapin, mapout,visited):
		if self.checkDep(visited):
			visited.add(self)
			i=self._i
			#print("visiting node number "+str(i))
			#update my LV_out set
			mapout[i]=set()  #new emptyset                
			for child in self._out:
				#print("seen a child, its id is "+str(child._i))
				mapout[i]=mapout[i].union(mapin[child._i])
			#update my LV_in set
			mapin[i]=(mapout[i].difference(self._kill)).union(self._gen)
			for child in self._out:
				child.do_dataflow_onestep(mapin, mapout,visited)
		else:
			pass
				
class BlockPROG(Block):
	pass

class BlockLABEL(Block):
	def __init__(self, s):
		Block.__init__(self)
		self._ins = Instru3A(s+':','','','')

class BlockADD(Block):
	def __init__(self, dr, sr1, sr2orimm7):
		Block.__init__(self)
		self._left = dr
		self._right = [sr1, sr2orimm7]
		self._ins = Instru3A("ADD",str(dr),str(sr1),str(sr2orimm7))
		self._kill.add(dr)
		self._gen.add(sr1)  #even if sr1 = dr!
		if not(RepresentsImm(sr2orimm7)):
			self._gen.add(sr2orimm7)
		if debug:
			print("Block "+str(self._i)) 
			print("KILL:",self._kill)
			print("GEN:",self._gen)
				
						
class BlockAND(Block):
	def __init__(self, dr, sr1, sr2orimm7):
		Block.__init__(self)
		self._left = dr
		self._right = [sr1, sr2orimm7]
		self._ins = Instru3A("AND",str(dr),str(sr1),str(sr2orimm7))
		# TODO: implement kill and gen initialization

						
class BlockNOT(Block):
	def __init__(self, dr,sr1):
		Block.__init__(self)
		self._left = dr
		self._right = [sr1]
		self._ins = Instru3A("NOT",str(dr),str(sr1),'')
		# TODO: implement kill and gen initialization

						
class BlockBR(Block):
	def __init__(self,s,label):
		Block.__init__(self)
		self._ins = Instru3A("BR"+s,str(label),'','')
		# TODO: implement kill and gen initialization

				
class BlockGOTO(Block):
	def __init__(self,label):
		Block.__init__(self)
		self._ins = Instru3A("BR",str(label),'','')
		# TODO: implement kill and gen initialization

### CFG MAIN CLASS
				
class CFG:
	def __init__(self, start):
		self._nbtmp = 0
		self._alltemps=[]
		self._nblabel=0
		self._start = start
		self._end = start
		self._mapin = {}  #will be map block -> set of variables
		self._mapout = {}
		self._mapdef =  {} #block : defined = killed vars in the block
		self._igraph = None #interference graph
				
	#Label Management.
	
	def newlabelIf(self):
		self._nblabel = self._nblabel+1
		return ("l_if_true_"+str(self._nblabel),"l_if_false_"+str(self._nblabel),"l_if_end_"+str(self._nblabel))

	def newlabelWhile(self):
		self._nblabel = self._nblabel+1
		return ("l_while_begin_"+str(self._nblabel),"l_while_end_"+str(self._nblabel))

	def newlabelCond(self):
		self._nblabel = self._nblabel+1
		return ("l_cond_true_"+str(self._nblabel),"l_cond_false_"+str(self._nblabel),"l_cond_end_"+str(self._nblabel))
	
	#### end of label management.
	
	#Temporaries Management.
	#all temps are prefixed by temp_
	def newtmp(self):
		self._nbtmp = self._nbtmp+1
		newvar='temp_'+str(self._nbtmp)
		self._alltemps.append(newvar)
		print(self._alltemps)
		return newvar

	def printTempList(self):
		print(self._alltemps)

	##### end of temp management.
				
	def setEnd(self, block):
		self._end = block
		return block

	def append(self, block):
		self._end.append(block)
		self._end = block
		return block

	def doDataflow(self):
		##### 
		# TODO : implement dataflow analysis ! 
		####
		raise Exception("Not Yet Implemented")

	def doInterfGraph(self):
		self._igraph = Graph()
		t = self._alltemps

		# TODO: implement interfere function
		raise Exception("Not Yet Implemented")

		return(self._igraph)

	def doAllocation(self):
		if not self._igraph:
			print("hum, the interference graph seems to be empty")
			exit(1)
		coloring=self._igraph.color(4)
		return(coloring)
				
		#print the 3address code / or allocated
	def printCode(self, filename, assignment=None):
		output = open(filename, 'w') if filename else sys.stdout
		print(';;Automatically generated code',file=output)
		self._start.printIns(output, set(), assignment)
		if output is not sys.stdout:
			output.close()

	def printDot(self, filename):
		graph = nx.DiGraph()
		self._start.printDot(graph, set())
		graph.graph['graph'] = dict()
		graph.graph['graph']['overlap'] = 'false'
		nx.drawing.nx_agraph.write_dot(graph, filename)
		gz.render('dot', 'pdf', filename) 
		gz.view(filename+'.pdf')


		
