from MuLexer import MuLexer
from MuParser import MuParser
from MuVisitor import MuVisitor
from CFGMuCodeGenVisitor import *

from APICodeLC3 import *
import networkx as nx
import graphviz as gz

from antlr4 import*
from sys import *
from os import *

from antlr4.tree.Trees import Trees

def main():
    inputname=argv[1]
    input_s=FileStream(inputname)
    lexer = MuLexer(input_s)
    stream = CommonTokenStream(lexer)
    parser = MuParser(stream)
    tree=parser.prog();     
    
    (hd,rest)=path.splitext(inputname)
    output_name=hd+".asm"
    output_name_dot=hd+".dot"
    print("Code will be generated in file "+output_name)
    
    #Instanciate a code generation Visitor, first argument is debug mode
    codeGenVisitor = CFGMuCodeGenVisitor(True,output_name,output_name_dot,parser);

    # build the cfg - does nothing for the moment!
    cfg = codeGenVisitor.visit(tree)

    #print allvars (for debug)
    cfg.printTempList()

    # 1) Build the CFG
    cfg.printDot(output_name_dot)

    # 2) Do the Liveness Analysis
    mapin,mapout = cfg.doDataflow()
    print("Live in map:")
    print(mapin)
    print("Live out map:")
    print(mapout)

    # 3) Do the Interference Graph
    igraph = cfg.doInterfGraph()

    # 4) Do the Coloration
    coloring = cfg.doAllocation()

    if coloring:
        print("coloring")
        print(coloring)
        igraph.print_dot(hd+"-colored.dot",coloring)
        cfg.printCode(output_name,coloring)
    else:
        print "Coloring failed"
        
if __name__ == '__main__':
    main()
    